from __future__ import annotations

import dataclasses
from abc import ABCMeta, abstractmethod
from collections import deque
from collections.abc import Iterable, Iterator, KeysView, MutableMapping
from types import MappingProxyType
from typing import Any

from itemadapter._imports import _scrapy_item_classes, attr
from itemadapter._json_schema import (
    _json_schema_from_attrs,
    _json_schema_from_dataclass,
    _json_schema_from_item_class,
    _json_schema_from_pydantic,
    _JsonSchemaState,
    _setdefault_attribute_docstrings_on_json_schema,
    _setdefault_attribute_types_on_json_schema,
)
from itemadapter.utils import (
    _get_pydantic_model_metadata,
    _get_pydantic_v1_model_metadata,
    _is_attrs_class,
    _is_pydantic_model,
    _is_pydantic_v1_model,
)

__all__ = [
    "AdapterInterface",
    "AttrsAdapter",
    "DataclassAdapter",
    "DictAdapter",
    "ItemAdapter",
    "PydanticAdapter",
    "ScrapyItemAdapter",
]


class AdapterInterface(MutableMapping, metaclass=ABCMeta):
    """Abstract Base Class for adapters.

    An adapter that handles a specific type of item should inherit from this
    class and implement the abstract methods defined here, plus the
    abtract methods inherited from the MutableMapping base class.
    """

    def __init__(self, item: Any) -> None:
        self.item = item

    @classmethod
    @abstractmethod
    def is_item_class(cls, item_class: type) -> bool:
        """Return True if the adapter can handle the given item class, False otherwise."""
        raise NotImplementedError

    @classmethod
    def is_item(cls, item: Any) -> bool:
        """Return True if the adapter can handle the given item, False otherwise."""
        return cls.is_item_class(item.__class__)

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        return MappingProxyType({})

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        """Return a list of fields defined for ``item_class``.
        If a class doesn't support fields, None is returned."""
        return None

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        return _json_schema_from_item_class(cls, item_class, _state)

    def get_field_meta(self, field_name: str) -> MappingProxyType:
        """Return metadata for the given field name, if available."""
        return self.get_field_meta_from_class(self.item.__class__, field_name)

    def field_names(self) -> KeysView:
        """Return a dynamic view of the item's field names."""
        return self.keys()  # type: ignore[return-value]


class _MixinAttrsDataclassAdapter:
    _fields_dict: dict
    item: Any

    def get_field_meta(self, field_name: str) -> MappingProxyType:
        return self._fields_dict[field_name].metadata

    def field_names(self) -> KeysView:
        return KeysView(self._fields_dict)

    def __getitem__(self, field_name: str) -> Any:
        if field_name in self._fields_dict:
            return getattr(self.item, field_name)
        raise KeyError(field_name)

    def __setitem__(self, field_name: str, value: Any) -> None:
        if field_name in self._fields_dict:
            setattr(self.item, field_name, value)
        else:
            raise KeyError(f"{self.item.__class__.__name__} does not support field: {field_name}")

    def __delitem__(self, field_name: str) -> None:
        if field_name in self._fields_dict:
            try:
                if hasattr(self.item, field_name):
                    delattr(self.item, field_name)
                else:
                    raise AttributeError
            except AttributeError as ex:
                raise KeyError(field_name) from ex
        else:
            raise KeyError(f"{self.item.__class__.__name__} does not support field: {field_name}")

    def __iter__(self) -> Iterator:
        return iter(attr for attr in self._fields_dict if hasattr(self.item, attr))

    def __len__(self) -> int:
        return len(list(iter(self)))


class AttrsAdapter(_MixinAttrsDataclassAdapter, AdapterInterface):
    def __init__(self, item: Any) -> None:
        super().__init__(item)
        if attr is None:
            raise RuntimeError("attr module is not available")
        # store a reference to the item's fields to avoid O(n) lookups and O(n^2) traversals
        self._fields_dict = attr.fields_dict(self.item.__class__)

    @classmethod
    def is_item(cls, item: Any) -> bool:
        return _is_attrs_class(item) and not isinstance(item, type)

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return _is_attrs_class(item_class)

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        if attr is None:
            raise RuntimeError("attr module is not available")
        try:
            return attr.fields_dict(item_class)[field_name].metadata
        except KeyError as ex:
            raise KeyError(f"{item_class.__name__} does not support field: {field_name}") from ex

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        if attr is None:
            raise RuntimeError("attr module is not available")
        return [a.name for a in attr.fields(item_class)]

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        _state = _state or _JsonSchemaState(adapter=cls, containers={item_class})
        return _json_schema_from_attrs(item_class, _state)


class DataclassAdapter(_MixinAttrsDataclassAdapter, AdapterInterface):
    def __init__(self, item: Any) -> None:
        super().__init__(item)
        # store a reference to the item's fields to avoid O(n) lookups and O(n^2) traversals
        self._fields_dict = {field.name: field for field in dataclasses.fields(self.item)}

    @classmethod
    def is_item(cls, item: Any) -> bool:
        return dataclasses.is_dataclass(item) and not isinstance(item, type)

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return dataclasses.is_dataclass(item_class)

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        for field in dataclasses.fields(item_class):
            if field.name == field_name:
                return field.metadata
        raise KeyError(f"{item_class.__name__} does not support field: {field_name}")

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        return [a.name for a in dataclasses.fields(item_class)]

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        _state = _state or _JsonSchemaState(adapter=cls, containers={item_class})
        return _json_schema_from_dataclass(item_class, _state)


class PydanticAdapter(AdapterInterface):
    item: Any

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return _is_pydantic_model(item_class) or _is_pydantic_v1_model(item_class)

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        try:
            try:
                return _get_pydantic_model_metadata(item_class, field_name)
            except AttributeError:
                return _get_pydantic_v1_model_metadata(item_class, field_name)
        except KeyError as ex:
            raise KeyError(f"{item_class.__name__} does not support field: {field_name}") from ex

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        try:
            return list(item_class.model_fields.keys())  # type: ignore[attr-defined]
        except AttributeError:
            return list(item_class.__fields__.keys())  # type: ignore[attr-defined]

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        return _json_schema_from_pydantic(cls, item_class, _state)

    def field_names(self) -> KeysView:
        try:
            return KeysView(self.item.__class__.model_fields)
        except AttributeError:
            return KeysView(self.item.__fields__)

    def __getitem__(self, field_name: str) -> Any:
        try:
            self.item.__class__.model_fields  # noqa: B018
        except AttributeError:
            if field_name in self.item.__fields__:
                return getattr(self.item, field_name)
        else:
            if field_name in self.item.__class__.model_fields:
                return getattr(self.item, field_name)
        raise KeyError(field_name)

    def __setitem__(self, field_name: str, value: Any) -> None:
        try:
            self.item.__class__.model_fields  # noqa: B018
        except AttributeError:
            if field_name in self.item.__fields__:
                setattr(self.item, field_name, value)
                return
        else:
            if field_name in self.item.__class__.model_fields:
                setattr(self.item, field_name, value)
                return
        raise KeyError(f"{self.item.__class__.__name__} does not support field: {field_name}")

    def __delitem__(self, field_name: str) -> None:
        try:
            self.item.__class__.model_fields  # noqa: B018
        except AttributeError as ex:
            if field_name in self.item.__fields__:
                try:
                    if hasattr(self.item, field_name):
                        delattr(self.item, field_name)
                        return
                    raise AttributeError from ex
                except AttributeError as ex2:
                    raise KeyError(field_name) from ex2
        else:
            if field_name in self.item.__class__.model_fields:
                try:
                    if hasattr(self.item, field_name):
                        delattr(self.item, field_name)
                        return
                    raise AttributeError
                except AttributeError as ex:
                    raise KeyError(field_name) from ex
        raise KeyError(f"{self.item.__class__.__name__} does not support field: {field_name}")

    def __iter__(self) -> Iterator:
        try:
            return iter(
                attr for attr in self.item.__class__.model_fields if hasattr(self.item, attr)
            )
        except AttributeError:
            return iter(attr for attr in self.item.__fields__ if hasattr(self.item, attr))

    def __len__(self) -> int:
        return len(list(iter(self)))


class _MixinDictScrapyItemAdapter:
    _fields_dict: dict
    item: Any

    def __getitem__(self, field_name: str) -> Any:
        return self.item[field_name]

    def __setitem__(self, field_name: str, value: Any) -> None:
        self.item[field_name] = value

    def __delitem__(self, field_name: str) -> None:
        del self.item[field_name]

    def __iter__(self) -> Iterator:
        return iter(self.item)

    def __len__(self) -> int:
        return len(self.item)


class DictAdapter(_MixinDictScrapyItemAdapter, AdapterInterface):
    @classmethod
    def is_item(cls, item: Any) -> bool:
        return isinstance(item, dict)

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return issubclass(item_class, dict)

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        return {"type": "object"}

    def field_names(self) -> KeysView:
        return KeysView(self.item)


class ScrapyItemAdapter(_MixinDictScrapyItemAdapter, AdapterInterface):
    @classmethod
    def is_item(cls, item: Any) -> bool:
        return isinstance(item, _scrapy_item_classes)

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return issubclass(item_class, _scrapy_item_classes)

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        return MappingProxyType(item_class.fields[field_name])  # type: ignore[attr-defined]

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        return list(item_class.fields.keys())  # type: ignore[attr-defined]

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        _state = _state or _JsonSchemaState(adapter=cls, containers={item_class})
        schema = super().get_json_schema(item_class, _state=_state)
        _setdefault_attribute_types_on_json_schema(schema, item_class, _state)
        _setdefault_attribute_docstrings_on_json_schema(schema, item_class)
        schema.pop("required", None)
        return schema

    def field_names(self) -> KeysView:
        return KeysView(self.item.fields)


class ItemAdapter(MutableMapping):
    """Wrapper class to interact with data container objects. It provides a common interface
    to extract and set data without having to take the object's type into account.
    """

    ADAPTER_CLASSES: Iterable[type[AdapterInterface]] = deque(
        [
            ScrapyItemAdapter,
            DictAdapter,
            DataclassAdapter,
            AttrsAdapter,
            PydanticAdapter,
        ]
    )

    def __init__(self, item: Any) -> None:
        for cls in self.ADAPTER_CLASSES:
            if cls.is_item(item):
                self.adapter = cls(item)
                break
        else:
            raise TypeError(f"No adapter found for objects of type: {type(item)} ({item})")

    @classmethod
    def is_item(cls, item: Any) -> bool:
        return any(adapter_class.is_item(item) for adapter_class in cls.ADAPTER_CLASSES)

    @classmethod
    def is_item_class(cls, item_class: type) -> bool:
        return any(
            adapter_class.is_item_class(item_class) for adapter_class in cls.ADAPTER_CLASSES
        )

    @classmethod
    def _get_adapter_class(cls, item_class: type) -> type[AdapterInterface]:
        for adapter_class in cls.ADAPTER_CLASSES:
            if adapter_class.is_item_class(item_class):
                return adapter_class
        raise TypeError(f"{item_class} is not a valid item class")

    @classmethod
    def get_field_meta_from_class(cls, item_class: type, field_name: str) -> MappingProxyType:
        adapter_class = cls._get_adapter_class(item_class)
        return adapter_class.get_field_meta_from_class(item_class, field_name)

    @classmethod
    def get_field_names_from_class(cls, item_class: type) -> list[str] | None:
        adapter_class = cls._get_adapter_class(item_class)
        return adapter_class.get_field_names_from_class(item_class)

    @classmethod
    def get_json_schema(
        cls, item_class: type, *, _state: _JsonSchemaState | None = None
    ) -> dict[str, Any]:
        _state = _state or _JsonSchemaState(adapter=cls, containers={item_class})
        adapter_class = cls._get_adapter_class(item_class)
        return adapter_class.get_json_schema(item_class, _state=_state)

    @property
    def item(self) -> Any:
        return self.adapter.item

    def __repr__(self) -> str:
        values = ", ".join([f"{key}={value!r}" for key, value in self.items()])
        return f"<{self.__class__.__name__} for {self.item.__class__.__name__}({values})>"

    def __getitem__(self, field_name: str) -> Any:
        return self.adapter.__getitem__(field_name)

    def __setitem__(self, field_name: str, value: Any) -> None:
        self.adapter.__setitem__(field_name, value)

    def __delitem__(self, field_name: str) -> None:
        self.adapter.__delitem__(field_name)

    def __iter__(self) -> Iterator:
        return self.adapter.__iter__()

    def __len__(self) -> int:
        return self.adapter.__len__()

    def get_field_meta(self, field_name: str) -> MappingProxyType:
        """Return metadata for the given field name."""
        return self.adapter.get_field_meta(field_name)

    def field_names(self) -> KeysView:
        """Return read-only key view with the names of all the defined fields for the item."""
        return self.adapter.field_names()

    def asdict(self) -> dict:
        """Return a dict object with the contents of the adapter. This works slightly different
        than calling `dict(adapter)`: it's applied recursively to nested items (if there are any).
        """
        return {key: self._asdict(value) for key, value in self.items()}

    @classmethod
    def _asdict(cls, obj: Any) -> Any:
        if isinstance(obj, dict):
            return {key: cls._asdict(value) for key, value in obj.items()}
        if isinstance(obj, (list, set, tuple)):
            return obj.__class__(cls._asdict(x) for x in obj)
        if isinstance(obj, cls):
            return obj.asdict()
        if cls.is_item(obj):
            return cls(obj).asdict()
        return obj



from __future__ import annotations

import ast
import dataclasses
import inspect
import operator
from collections.abc import Iterator, Mapping, Sequence
from collections.abc import Set as AbstractSet
from copy import copy
from enum import Enum
from textwrap import dedent
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    runtime_checkable,
)

from ._imports import PydanticUndefined, PydanticV1Undefined, attr
from .utils import _is_pydantic_model

if TYPE_CHECKING:
    from types import MappingProxyType

    from .adapter import AdapterInterface, ItemAdapter


SIMPLE_TYPES = {
    type(None): "null",
    bool: "boolean",
    int: "integer",
    float: "number",
    str: "string",
}


@dataclasses.dataclass
class _JsonSchemaState:
    adapter: type[ItemAdapter | AdapterInterface]
    """ItemAdapter class or AdapterInterface implementation used on the initial
    get_json_schema() call.

    On types for which adapter.is_item_class() returns True,
    adapter.get_json_schema() is used to get the corresponding, nested JSON
    Schema.
    """
    containers: set[type] = dataclasses.field(default_factory=set)
    """Used to keep track of item classes that are being processed, to avoid
    recursion."""


def dedupe_types(types: Sequence[type]) -> list[type]:
    seen = set()
    result = []
    for t in types:
        key = float if t in (int, float) else t
        if key not in seen:
            seen.add(key)
            result.append(t)
    return result


def update_prop_from_union(prop: dict[str, Any], prop_type: Any, state: _JsonSchemaState) -> None:
    prop_types = dedupe_types(get_args(prop_type))
    simple_types = [v for k, v in SIMPLE_TYPES.items() if k in prop_types]
    complex_types = sorted([t for t in prop_types if t not in SIMPLE_TYPES])  # type: ignore[type-var]
    if not complex_types:
        prop.setdefault("type", simple_types)
        return
    new_any_of: list[dict[str, Any]] = []
    any_of = prop.setdefault("anyOf", new_any_of)
    if any_of is not new_any_of:
        return
    any_of.append({"type": simple_types if len(simple_types) > 1 else simple_types[0]})
    for complex_type in complex_types:
        complex_prop: dict[str, Any] = {}
        update_prop_from_type(complex_prop, complex_type, state)
        any_of.append(complex_prop)


@runtime_checkable
class ArrayProtocol(Protocol):
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def __contains__(self, item: Any) -> bool: ...


@runtime_checkable
class ObjectProtocol(Protocol):  # noqa: PLW1641
    def __getitem__(self, key: str) -> Any: ...
    def __iter__(self) -> Iterator[str]: ...
    def __len__(self) -> int: ...
    def __contains__(self, key: str) -> bool: ...
    def keys(self): ...
    def items(self): ...
    def values(self): ...
    def get(self, key: str, default: Any = ...): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...


INVALID_PATTERN_SUBSTRINGS = [
    "(?P<",  # named groups
    "(?<=",  # lookbehind
    "(?<!",  # negative lookbehind
    "(?>",  # atomic group
    "\\A",  # start of string
    "\\Z",  # end of string
    "(?i)",  # inline flags (case-insensitive, etc.)
    "(?m)",  # multiline
    "(?s)",  # dotall
    "(?x)",  # verbose
    "(?#",  # comments
]


def is_valid_pattern(pattern: str) -> bool:
    # https://ecma-international.org/publications-and-standards/standards/ecma-262/
    #
    # Note: We allow word boundaries (\b, \B) in patterns even thought there is
    # a difference in behavior: in Python, they work with Unicode; in JSON
    # Schema, they only work with ASCII.
    return not any(sub in pattern for sub in INVALID_PATTERN_SUBSTRINGS)


def array_type(type_hint):
    """Given the type hint of a Python type that maps to a JSON Schema array,
    such as a list, a tuple or a set, return the type of the items in that
    array."""
    args = get_args(type_hint)
    if not args:
        return Any
    if args[-1] is Ellipsis:
        args = args[:-1]
    unique_args = set(args)
    if len(unique_args) == 1:
        return next(iter(unique_args))
    return Union[tuple(unique_args)]


def update_prop_from_pattern(prop: dict[str, Any], pattern: str) -> None:
    if is_valid_pattern(pattern):
        prop.setdefault("pattern", pattern)


try:
    from types import UnionType
except ImportError:  # Python < 3.10
    UNION_TYPES: set[Any] = {Union}
else:
    UNION_TYPES = {Union, UnionType}


def update_prop_from_origin(
    prop: dict[str, Any], origin: Any, prop_type: Any, state: _JsonSchemaState
) -> None:
    if isinstance(origin, type):
        if issubclass(origin, (Sequence, AbstractSet)):
            prop.setdefault("type", "array")
            if issubclass(origin, AbstractSet):
                prop.setdefault("uniqueItems", True)
            had_items = "items" in prop
            items = prop.setdefault("items", {})
            item_type = array_type(prop_type)
            update_prop_from_type(items, item_type, state)
            if not items and not had_items:
                del prop["items"]
            return
        if issubclass(origin, Mapping):
            prop.setdefault("type", "object")
            args = get_args(prop_type)
            if args:
                assert len(args) == 2
                value_type = args[1]
                props = prop.setdefault("additionalProperties", {})
                update_prop_from_type(props, value_type, state)
            return
    if origin in UNION_TYPES:
        update_prop_from_union(prop, prop_type, state)


def update_prop_from_type(prop: dict[str, Any], prop_type: Any, state: _JsonSchemaState) -> None:
    if (origin := get_origin(prop_type)) is not None:
        update_prop_from_origin(prop, origin, prop_type, state)
        return
    if isinstance(prop_type, type):
        if state.adapter.is_item_class(prop_type):
            if prop_type in state.containers:
                prop.setdefault("type", "object")
                return
            state.containers.add(prop_type)
            subschema = state.adapter.get_json_schema(
                prop_type,
                _state=state,
            )
            state.containers.remove(prop_type)
            for k, v in subschema.items():
                prop.setdefault(k, v)
            return
        if issubclass(prop_type, Enum):
            values = [item.value for item in prop_type]
            value_types = tuple({type(v) for v in values})
            prop_type = value_types[0] if len(value_types) == 1 else Union[value_types]
            update_prop_from_type(prop, prop_type, state)
            prop.setdefault("enum", values)
            return
        if not issubclass(prop_type, str):
            if isinstance(prop_type, ObjectProtocol):
                prop.setdefault("type", "object")
                return
            if isinstance(prop_type, ArrayProtocol):
                prop.setdefault("type", "array")
                if issubclass(prop_type, AbstractSet):
                    prop.setdefault("uniqueItems", True)
                return
    json_schema_type = SIMPLE_TYPES.get(prop_type)
    if json_schema_type is not None:
        prop.setdefault("type", json_schema_type)


def _setdefault_attribute_types_on_json_schema(
    schema: dict[str, Any], item_class: type, state: _JsonSchemaState
) -> None:
    """Inspect the type hints of the class attributes of the item class and,
    for any matching JSON Schema property that has no type set, set the type
    based on the type hint."""
    props = schema.get("properties", {})
    attribute_type_hints = get_type_hints(item_class)
    for prop_name, prop in props.items():
        if prop_name not in attribute_type_hints:
            continue
        prop_type = attribute_type_hints[prop_name]
        update_prop_from_type(prop, prop_type, state)


def iter_docstrings(item_class: type, attr_names: AbstractSet[str]) -> Iterator[tuple[str, str]]:
    try:
        source = inspect.getsource(item_class)
    except (OSError, TypeError):
        return
    tree = ast.parse(dedent(source))
    class_node = None
    for node in tree.body:
        if isinstance(node, ast.ClassDef) and node.name == item_class.__name__:
            class_node = node
            break
    if class_node is None:  # pragma: no cover
        # This can be reproduced with the doctests of the README, but the
        # coverage data does not seem to include those.
        return
    assert isinstance(class_node, ast.ClassDef)
    for child in ast.iter_child_nodes(class_node):
        if isinstance(child, ast.Assign) and isinstance(child.targets[0], ast.Name):
            attr_name = child.targets[0].id
        elif isinstance(child, ast.AnnAssign) and isinstance(child.target, ast.Name):
            attr_name = child.target.id
        else:
            continue
        if attr_name not in attr_names:
            continue
        next_idx = class_node.body.index(child) + 1
        if next_idx >= len(class_node.body):
            continue
        next_node = class_node.body[next_idx]
        if (
            isinstance(next_node, ast.Expr)
            and isinstance(next_node.value, ast.Constant)
            and isinstance(next_node.value.value, str)
        ):
            yield attr_name, next_node.value.value


def get_inherited_attr_docstring(item_class: type, attr_name: str) -> str | None:
    """Recursively search the MRO for a docstring for the given attribute
    name."""
    for cls in item_class.__mro__:
        for name, doc in iter_docstrings(cls, {attr_name}):
            if name == attr_name:
                return doc
    return None


def _setdefault_attribute_docstrings_on_json_schema(
    schema: dict[str, Any], item_class: type
) -> None:
    """Inspect the docstrings after each class attribute of the item class and
    its bases and, for any matching JSON Schema property that has no
    description set, set the description to the contents of the docstring."""
    props = schema.get("properties", {})
    attr_names = set(props)
    if not attr_names:
        return
    for attr_name in attr_names:
        prop = props.setdefault(attr_name, {})
        if "description" not in prop:
            doc = get_inherited_attr_docstring(item_class, attr_name)
            if doc:
                prop["description"] = inspect.cleandoc(doc)


def base_json_schema_from_item_class(item_class: type) -> dict[str, Any]:
    json_schema_extra = getattr(item_class, "__json_schema_extra__", {})
    schema = copy(json_schema_extra)
    schema.setdefault("type", "object")
    schema.setdefault("additionalProperties", False)
    return schema


def _json_schema_from_item_class(
    adapter: type[AdapterInterface], item_class: type, state: _JsonSchemaState | None = None
) -> dict[str, Any]:
    state = state or _JsonSchemaState(adapter=adapter, containers={item_class})
    schema = base_json_schema_from_item_class(item_class)
    fields_meta = {
        field_name: adapter.get_field_meta_from_class(item_class, field_name)
        for field_name in adapter.get_field_names_from_class(item_class) or ()
    }
    if not fields_meta:
        return schema
    schema["properties"] = {
        field_name: copy(field_meta.get("json_schema_extra", {}))
        for field_name, field_meta in fields_meta.items()
    }
    required = [
        field_name
        for field_name, field_data in schema["properties"].items()
        if "default" not in field_data
    ]
    if required:
        schema.setdefault("required", required)
    return schema


def update_required_fields(
    schema: dict[str, Any], optional_fields: set[str] | None = None
) -> None:
    optional_fields = optional_fields or set()
    if "required" in schema:
        return
    required = [
        field
        for field, metadata in schema["properties"].items()
        if field not in optional_fields and "default" not in metadata
    ]
    if required:
        schema["required"] = required


def _json_schema_from_attrs(item_class: type, state: _JsonSchemaState) -> dict[str, Any]:
    schema = base_json_schema_from_item_class(item_class)
    fields = attr.fields(item_class)
    if not fields:
        return schema

    from attr import resolve_types

    resolve_types(item_class)  # Ensure field.type annotations are resolved

    schema["properties"] = {
        field.name: copy(field.metadata.get("json_schema_extra", {})) for field in fields
    }
    default_factory_fields: set[str] = set()
    for field in fields:
        prop = schema["properties"][field.name]
        _update_attrs_prop(prop, field, state, default_factory_fields)
    update_required_fields(schema, default_factory_fields)
    _setdefault_attribute_docstrings_on_json_schema(schema, item_class)
    return schema


def _update_attrs_prop(
    prop: dict[str, Any],
    field: attr.Attribute,
    state: _JsonSchemaState,
    default_factory_fields: set[str],
) -> None:
    update_prop_from_type(prop, field.type, state)
    if isinstance(field.default, attr.Factory):
        default_factory_fields.add(field.name)
    elif field.default is not attr.NOTHING:
        prop.setdefault("default", field.default)
    _update_attrs_prop_validation(prop, field)


ATTRS_NUMBER_VALIDATORS = {
    operator.ge: "minimum",
    operator.gt: "exclusiveMinimum",
    operator.le: "maximum",
    operator.lt: "exclusiveMaximum",
}


def _update_attrs_prop_validation(
    prop: dict[str, Any],
    field: attr.Attribute,
) -> None:
    if not field.validator:
        return
    if type(field.validator).__name__ == "_AndValidator":
        validators = field.validator._validators
    else:
        validators = [field.validator]
    for validator in validators:
        validator_type_name = type(validator).__name__
        if validator_type_name == "_NumberValidator":
            key = ATTRS_NUMBER_VALIDATORS.get(validator.compare_func)
            if not key:  # pragma: no cover
                continue
            prop.setdefault(key, validator.bound)
        elif validator_type_name == "_InValidator":
            prop.setdefault("enum", list(validator.options))
        elif validator_type_name == "_MinLengthValidator":
            key = "minLength" if field.type is str else "minItems"
            prop.setdefault(key, validator.min_length)
        elif validator_type_name == "_MaxLengthValidator":
            key = "maxLength" if field.type is str else "maxItems"
            prop.setdefault(key, validator.max_length)
        elif validator_type_name == "_MatchesReValidator":
            pattern_obj = getattr(validator, "pattern", None) or validator.regex
            update_prop_from_pattern(prop, pattern_obj.pattern)


def _json_schema_from_dataclass(item_class: type, state: _JsonSchemaState) -> dict[str, Any]:
    schema = base_json_schema_from_item_class(item_class)
    fields = dataclasses.fields(item_class)
    resolved_field_types = get_type_hints(item_class)
    default_factory_fields = set()
    if fields:
        schema["properties"] = {
            field.name: copy(field.metadata.get("json_schema_extra", {})) for field in fields
        }
        for field in fields:
            prop = schema["properties"][field.name]
            field_type = resolved_field_types.get(field.name)
            if field_type is not None:
                update_prop_from_type(prop, field_type, state)
            if field.default_factory is not dataclasses.MISSING:
                default_factory_fields.add(field.name)
            elif field.default is not dataclasses.MISSING:
                prop.setdefault("default", field.default)
        update_required_fields(schema, default_factory_fields)
    _setdefault_attribute_docstrings_on_json_schema(schema, item_class)
    return schema


def _json_schema_from_pydantic(
    adapter: type[AdapterInterface], item_class: type, state: _JsonSchemaState | None = None
) -> dict[str, Any]:
    state = state or _JsonSchemaState(adapter=adapter, containers={item_class})
    if not _is_pydantic_model(item_class):
        return _json_schema_from_pydantic_v1(adapter, item_class, state)
    schema = copy(
        item_class.model_config.get("json_schema_extra", {})  # type: ignore[attr-defined]
    )
    extra = item_class.model_config.get("extra")  # type: ignore[attr-defined]
    schema.setdefault("type", "object")
    if extra == "forbid":
        schema.setdefault("additionalProperties", False)
    fields = {
        name: adapter.get_field_meta_from_class(item_class, name)
        for name in adapter.get_field_names_from_class(item_class) or ()
    }
    if not fields:
        return schema
    schema["properties"] = {
        name: copy(metadata.get("json_schema_extra", {})) for name, metadata in fields.items()
    }
    default_factory_fields: set[str] = set()
    for name, metadata in fields.items():
        prop = schema["properties"][name]
        _update_pydantic_prop(prop, name, metadata, state, default_factory_fields)
    update_required_fields(schema, default_factory_fields)
    _setdefault_attribute_docstrings_on_json_schema(schema, item_class)
    return schema


def _update_pydantic_prop(
    prop: dict[str, Any],
    name: str,
    metadata: MappingProxyType,
    _state: _JsonSchemaState,
    default_factory_fields: set[str],
) -> None:
    if "annotation" in metadata:
        field_type = metadata["annotation"]
        if field_type is not None:
            update_prop_from_type(prop, field_type, _state)
    if "default_factory" in metadata:
        default_factory_fields.add(name)
    elif "default" in metadata and metadata["default"] is not PydanticUndefined:
        prop.setdefault("default", metadata["default"])
    if "metadata" in metadata:
        _update_pydantic_prop_validation(prop, metadata["metadata"], field_type)
    for metadata_key, json_schema_field in (
        ("title", "title"),
        ("description", "description"),
        ("examples", "examples"),
    ):
        if metadata_key in metadata:
            prop.setdefault(json_schema_field, metadata[metadata_key])
    if "deprecated" in metadata:
        prop.setdefault("deprecated", bool(metadata["deprecated"]))


def _update_pydantic_prop_validation(
    prop: dict[str, Any],
    metadata: Sequence[Any],
    field_type: type,
) -> None:
    for metadata_item in metadata:
        metadata_item_type = type(metadata_item).__name__
        if metadata_item_type == "_PydanticGeneralMetadata":
            if "pattern" in metadata_item.__dict__:
                pattern = metadata_item.__dict__["pattern"]
                update_prop_from_pattern(prop, pattern)
        elif metadata_item_type == "MinLen":
            key = "minLength" if field_type is str else "minItems"
            prop.setdefault(key, metadata_item.min_length)
        elif metadata_item_type == "MaxLen":
            key = "maxLength" if field_type is str else "maxItems"
            prop.setdefault(key, metadata_item.max_length)
        else:
            for metadata_key, json_schema_field in (
                ("ge", "minimum"),
                ("gt", "exclusiveMinimum"),
                ("le", "maximum"),
                ("lt", "exclusiveMaximum"),
            ):
                if metadata_item_type == metadata_key.capitalize():
                    prop.setdefault(json_schema_field, getattr(metadata_item, metadata_key))


def _json_schema_from_pydantic_v1(
    adapter: type[AdapterInterface], item_class: type, state: _JsonSchemaState
) -> dict[str, Any]:
    schema = copy(
        getattr(item_class.Config, "schema_extra", {})  # type: ignore[attr-defined]
    )
    extra = getattr(item_class.Config, "extra", None)  # type: ignore[attr-defined]
    schema.setdefault("type", "object")
    if extra == "forbid":
        schema.setdefault("additionalProperties", False)
    fields = {
        name: adapter.get_field_meta_from_class(item_class, name)
        for name in adapter.get_field_names_from_class(item_class) or ()
    }
    if not fields:
        return schema
    schema["properties"] = {
        name: copy(metadata.get("json_schema_extra", {})) for name, metadata in fields.items()
    }
    default_factory_fields: set[str] = set()
    field_type_hints = get_type_hints(item_class)
    for name, metadata in fields.items():
        prop = schema["properties"][name]
        _update_pydantic_v1_prop(
            prop, name, metadata, field_type_hints, default_factory_fields, state
        )
    update_required_fields(schema, default_factory_fields)
    _setdefault_attribute_docstrings_on_json_schema(schema, item_class)
    return schema


def _update_pydantic_v1_prop(  # pylint: disable=too-many-positional-arguments,too-many-arguments
    prop: dict[str, Any],
    name: str,
    metadata: Mapping[str, Any],
    field_type_hints: dict[str, Any],
    default_factory_fields: set[str],
    state: _JsonSchemaState,
) -> None:
    field_type = field_type_hints[name]
    if field_type is not None:
        update_prop_from_type(prop, field_type, state)
    if "default_factory" in metadata:
        default_factory_fields.add(name)
    elif "default" in metadata and metadata["default"] not in (
        Ellipsis,
        PydanticV1Undefined,
    ):
        prop.setdefault("default", metadata["default"])
    for metadata_key, json_schema_field in (
        ("title", "title"),
        ("description", "description"),
        ("examples", "examples"),
        ("ge", "minimum"),
        ("gt", "exclusiveMinimum"),
        ("le", "maximum"),
        ("lt", "exclusiveMaximum"),
    ):
        if metadata_key in metadata:
            prop.setdefault(json_schema_field, metadata[metadata_key])
    for prefix in ("min", "max"):
        if f"{prefix}_length" in metadata:
            key = f"{prefix}Length" if field_type is str else f"{prefix}Items"
            prop.setdefault(key, metadata[f"{prefix}_length"])
        elif f"{prefix}_items" in metadata:
            prop.setdefault(f"{prefix}Items", metadata[f"{prefix}_items"])
    for metadata_key in ("pattern", "regex"):
        if metadata_key in metadata:
            pattern = metadata[metadata_key]
            update_prop_from_pattern(prop, pattern)
            break
    if "deprecated" in metadata:
        prop.setdefault("deprecated", bool(metadata["deprecated"]))



from __future__ import annotations

from typing import Any

# attempt the following imports only once,
# to be imported from itemadapter's submodules


_scrapy_item_classes: tuple
scrapy: Any

try:
    import scrapy
except ImportError:
    _scrapy_item_classes = ()
    scrapy = None
else:
    try:
        # handle deprecated base classes
        _base_item_cls = getattr(
            scrapy.item,
            "_BaseItem",
            scrapy.item.BaseItem,
        )
    except AttributeError:
        _scrapy_item_classes = (scrapy.item.Item,)
    else:
        _scrapy_item_classes = (scrapy.item.Item, _base_item_cls)

attr: Any
try:
    import attr
except ImportError:
    attr = None

pydantic_v1: Any = None
pydantic: Any = None

try:
    import pydantic
except ImportError:  # No pydantic
    pass
else:
    try:
        import pydantic.v1 as pydantic_v1
    except ImportError:  # Pydantic <1.10.17
        pydantic_v1 = pydantic
        pydantic = None
    else:  # Pydantic 1.10.17+
        if not hasattr(pydantic.BaseModel, "model_fields"):  # Pydantic <2
            pydantic_v1 = pydantic
            pydantic = None

try:
    from pydantic.v1.fields import Undefined as PydanticV1Undefined
    from pydantic_core import PydanticUndefined
except ImportError:  # < Pydantic 2.0
    try:
        from pydantic.fields import (  # type: ignore[attr-defined,no-redef]
            Undefined as PydanticUndefined,
        )
        from pydantic.fields import (  # type: ignore[attr-defined,no-redef]
            Undefined as PydanticV1Undefined,
        )
    except ImportError:
        PydanticUndefined = PydanticV1Undefined = None  # type: ignore[assignment]



from __future__ import annotations

from types import MappingProxyType
from typing import Any

from itemadapter._imports import (
    PydanticUndefined,
    PydanticV1Undefined,
    attr,
    pydantic,
    pydantic_v1,
)

__all__ = ["get_field_meta_from_class", "is_item"]


def _is_attrs_class(obj: Any) -> bool:
    if attr is None:
        return False
    return attr.has(obj)


def _is_pydantic_model(obj: Any) -> bool:
    if pydantic is None:
        return False
    return issubclass(obj, pydantic.BaseModel)


def _is_pydantic_v1_model(obj: Any) -> bool:
    if pydantic_v1 is None:
        return False
    return issubclass(obj, pydantic_v1.BaseModel)


def _get_pydantic_model_metadata(item_model: Any, field_name: str) -> MappingProxyType:
    metadata = {}
    field = item_model.model_fields[field_name]

    for attribute in [
        "alias_priority",
        "alias",
        "allow_inf_nan",
        "annotation",
        "coerce_numbers_to_str",
        "decimal_places",
        "default_factory",
        "deprecated",
        "description",
        "discriminator",
        "examples",
        "exclude",
        "fail_fast",
        "field_title_generator",
        "frozen",
        "ge",
        "gt",
        "init_var",
        "init",
        "json_schema_extra",
        "kw_only",
        "le",
        "lt",
        "max_digits",
        "max_length",
        "min_length",
        "multiple_of",
        "pattern",
        "repr",
        "serialization_alias",
        "strict",
        "title",
        "union_mode",
        "validate_default",
        "validation_alias",
    ]:
        if hasattr(field, attribute) and (value := getattr(field, attribute)) is not None:
            metadata[attribute] = value

    for attribute, default_value in [
        ("default", PydanticUndefined),
        ("metadata", []),
    ]:
        if hasattr(field, attribute) and (value := getattr(field, attribute)) != default_value:
            metadata[attribute] = value

    return MappingProxyType(metadata)


def _get_pydantic_v1_model_metadata(item_model: Any, field_name: str) -> MappingProxyType:
    metadata = {}
    field = item_model.__fields__[field_name]
    field_info = field.field_info

    for attribute in [
        "alias",
        "const",
        "description",
        "ge",
        "gt",
        "le",
        "lt",
        "max_items",
        "max_length",
        "min_items",
        "min_length",
        "multiple_of",
        "regex",
        "title",
    ]:
        value = getattr(field_info, attribute)
        if value is not None:
            metadata[attribute] = value

    if (value := field_info.default) not in (PydanticV1Undefined, Ellipsis):
        metadata["default"] = value

    if value := field.default_factory is not None:
        metadata["default_factory"] = value

    if not field_info.allow_mutation:
        metadata["allow_mutation"] = field_info.allow_mutation
    metadata.update(field_info.extra)

    return MappingProxyType(metadata)


def is_item(obj: Any) -> bool:
    """Return True if the given object belongs to one of the supported types, False otherwise.

    Alias for ItemAdapter.is_item
    """
    from itemadapter.adapter import ItemAdapter

    return ItemAdapter.is_item(obj)


def get_field_meta_from_class(item_class: type, field_name: str) -> MappingProxyType:
    """Return a read-only mapping with metadata for the given field name, within the given
    item class. If there is no metadata for the field, or the item class does not support
    field metadata, an empty object is returned.

    Field metadata is taken from different sources, depending on the item type:
    * scrapy.item.Item: corresponding scrapy.item.Field object
    * dataclass items: "metadata" attribute for the corresponding field
    * attrs items: "metadata" attribute for the corresponding field
    * pydantic models: corresponding pydantic.field.FieldInfo/ModelField object

    The returned value is an instance of types.MappingProxyType, i.e. a dynamic read-only view
    of the original mapping, which gets automatically updated if the original mapping changes.
    """

    from itemadapter.adapter import ItemAdapter

    return ItemAdapter.get_field_meta_from_class(item_class, field_name)
